@@ -1,10 +1,10 @@
 import distutils.command.bdist_msi
 import distutils.errors
-import distutils.util
 import importlib
 import msilib
 import os
 import re
+import sysconfig
 
 __all__ = ["bdist_msi"]
 
@@ -28,16 +28,19 @@ class bdist_msi(distutils.command.bdist_msi.bdist_msi):
         (
             "data=",
             None,
-            "dictionary of data indexed by table name, and each value is a tuple to include in table",
+            "dictionary of data indexed by table name, and each value is a "
+            "tuple to include in table",
         ),
         (
             "summary-data=",
             None,
-            'Dictionary of data to include in msi summary data stream.  Allowed keys are "author", "comments", "keywords".',
+            "Dictionary of data to include in msi summary data stream. "
+            'Allowed keys are "author", "comments", "keywords".',
         ),
         ("product-code=", None, "product code to use"),
         ("install-icon=", None, "icon path to add/remove programs "),
         ("all-users=", None, "installation for all users (or just me)"),
+        ("extensions=", None, "Extensions for which to register Verbs"),
     ]
     x = y = 50
     width = 370
@@ -45,6 +48,14 @@ class bdist_msi(distutils.command.bdist_msi.bdist_msi):
     title = "[ProductName] Setup"
     modeless = 1
     modal = 3
+    _binary_columns = {
+        "Binary": 1,
+        "Icon": 1,
+        "Patch": 4,
+        "SFPCatalog": 1,
+        "MsiDigitalCertificate": 1,
+        "MsiPatchHeaders": 1,
+    }
 
     def add_config(self, fullname):
         if self.add_to_path:
@@ -99,18 +110,18 @@ class bdist_msi(distutils.command.bdist_msi.bdist_msi):
         )
         for index, executable in enumerate(self.distribution.executables):
             if (
-                executable.shortcutName is not None
-                and executable.shortcutDir is not None
+                executable.shortcut_name is not None
+                and executable.shortcut_dir is not None
             ):
-                baseName = os.path.basename(executable.targetName)
+                baseName = os.path.basename(executable.target_name)
                 msilib.add_data(
                     self.db,
                     "Shortcut",
                     [
                         (
                             "S_APP_%s" % index,
-                            executable.shortcutDir,
-                            executable.shortcutName,
+                            executable.shortcut_dir,
+                            executable.shortcut_name,
                             "TARGETDIR",
                             "[TARGETDIR]%s" % baseName,
                             None,
@@ -119,11 +130,17 @@ class bdist_msi(distutils.command.bdist_msi.bdist_msi):
                             None,
                             None,
                             None,
-                            None,
+                            "TARGETDIR",
                         )
                     ],
                 )
         for tableName, data in self.data.items():
+            col = self._binary_columns.get(tableName)
+            if col is not None:
+                data = [
+                    (*row[:col], msilib.Binary(row[col]), *row[col + 1 :])
+                    for row in data
+                ]
             msilib.add_data(self.db, tableName, data)
 
         # If provided, add data to MSI's summary information stream
@@ -294,14 +311,26 @@ class bdist_msi(distutils.command.bdist_msi.bdist_msi):
         while todo:
             dir = todo.pop()
             for file in os.listdir(dir.absolute):
-                if os.path.isdir(os.path.join(dir.absolute, file)):
+                sep_comp = self.separate_components.get(
+                    os.path.relpath(
+                        os.path.join(dir.absolute, file),
+                        self.bdist_dir,
+                    )
+                )
+                if sep_comp is not None:
+                    restore_component = dir.component
+                    dir.start_component(
+                        component=sep_comp,
+                        flags=0,
+                        feature=f,
+                        keyfile=file,
+                    )
+                    dir.add_file(file)
+                    dir.component = restore_component
+                elif os.path.isdir(os.path.join(dir.absolute, file)):
+                    short_file = dir.make_short(file)
                     newDir = msilib.Directory(
-                        db,
-                        cab,
-                        dir,
-                        file,
-                        file,
-                        "{}|{}".format(dir.make_short(file), file),
+                        db, cab, dir, file, file, f"{short_file}|{file}"
                     )
                     todo.append(newDir)
                 else:
@@ -744,12 +773,18 @@ class bdist_msi(distutils.command.bdist_msi.bdist_msi):
         )
         button.event("EndDialog", "Exit")
 
+    def _append_to_data(self, table, *line):
+        rows = self.data.setdefault(table, [])
+        line = tuple(line)
+        if line not in rows:
+            rows.append(line)
+
     def finalize_options(self):
         distutils.command.bdist_msi.bdist_msi.finalize_options(self)
         name = self.distribution.get_name()
         fullname = self.distribution.get_fullname()
         if self.initial_target_dir is None:
-            if distutils.util.get_platform() == "win-amd64":
+            if sysconfig.get_platform() == "win-amd64":
                 programFilesFolder = "ProgramFiles64Folder"
             else:
                 programFilesFolder = "ProgramFilesFolder"
@@ -759,7 +794,7 @@ class bdist_msi(distutils.command.bdist_msi.bdist_msi):
         if self.target_name is None:
             self.target_name = fullname
         if not self.target_name.lower().endswith(".msi"):
-            platform = distutils.util.get_platform().replace("win-", "")
+            platform = sysconfig.get_platform().replace("win-", "")
             self.target_name = f"{self.target_name}-{platform}.msi"
         if not os.path.isabs(self.target_name):
             self.target_name = os.path.join(self.dist_dir, self.target_name)
@@ -771,6 +806,104 @@ class bdist_msi(distutils.command.bdist_msi.bdist_msi):
             self.data = {}
         if not isinstance(self.summary_data, dict):
             self.summary_data = {}
+        self.separate_components = {}
+        for n, executable in enumerate(self.distribution.executables):
+            base_name = os.path.basename(executable.target_name)
+            # Trying to make these names unique from any directory name
+            self.separate_components[base_name] = msilib.make_id(
+                f"_cx_executable{n}_{executable}"
+            )
+        if self.extensions is None:
+            self.extensions = []
+        for extension in self.extensions:
+            try:
+                ext = extension["extension"]
+                verb = extension["verb"]
+                executable = extension["executable"]
+            except KeyError:
+                raise ValueError(
+                    "Each extension must have at least extension, verb, "
+                    "and executable"
+                )
+            try:
+                component = self.separate_components[executable]
+            except KeyError:
+                raise ValueError(
+                    "Executable must be the base target name of one of the "
+                    "distribution's executables"
+                )
+            progid = msilib.make_id(
+                "{}.{}.{}".format(
+                    self.distribution.get_name(),
+                    os.path.splitext(executable)[0],
+                    self.distribution.get_version(),
+                )
+            )
+            mime = extension.get("mime", None)
+            # "%1" a better default for argument?
+            argument = extension.get("argument", None)
+            context = extension.get("context", f"{fullname} {verb}")
+            # Add rows via self.data to safely ignore duplicates
+            self._append_to_data(
+                "ProgId",
+                progid,
+                None,
+                None,
+                self.distribution.get_description(),
+                None,
+                None,
+            )
+            self._append_to_data(
+                "Extension",
+                ext,
+                component,
+                progid,
+                mime,
+                "default",
+            )
+            self._append_to_data(
+                "Verb",
+                ext,
+                verb,
+                0,
+                context,
+                argument,
+            )
+            if mime is not None:
+                self._append_to_data(
+                    "MIME",
+                    mime,
+                    ext,
+                    "None",
+                )
+            # Registry entries that allow proper display of the app in menu
+            self._append_to_data(
+                "Registry",
+                f"{progid}-name",
+                -1,
+                fr"Software\Classes\{progid}",
+                "FriendlyAppName",
+                self.distribution.get_name(),
+                component,
+            )
+            self._append_to_data(
+                "Registry",
+                f"{progid}-verb-{verb}",
+                -1,
+                fr"Software\Classes\{progid}\shell\{verb}",
+                "FriendlyAppName",
+                self.distribution.get_name(),
+                component,
+            )
+            self._append_to_data(
+                "Registry",
+                f"{progid}-author",
+                -1,
+                fr"Software\Classes\{progid}\Application",
+                "ApplicationCompany",
+                self.distribution.get_author(),
+                component,
+            )
 
     def initialize_options(self):
         distutils.command.bdist_msi.bdist_msi.initialize_options(self)
@@ -785,6 +918,7 @@ class bdist_msi(distutils.command.bdist_msi.bdist_msi):
         self.summary_data = None
         self.install_icon = None
         self.all_users = False
+        self.extensions = None
 
     def run(self):
         if not self.skip_build:
@@ -804,7 +938,7 @@ class bdist_msi(distutils.command.bdist_msi.bdist_msi):
         author = metadata.author or metadata.maintainer or "UNKNOWN"
         version = metadata.get_version()
         sversion = ".".join(
-            [str(x) for x in distutils.version.LooseVersion(version).version]
+            str(x) for x in distutils.version.LooseVersion(version).version
         )
 
         # msilib is reloaded in order to reset the "_directories" global member
@@ -842,8 +976,9 @@ class bdist_msi(distutils.command.bdist_msi.bdist_msi):
                 self.bdist_dir, dry_run=self.dry_run
             )
 
-        # Cause the MSI file to be released.  Without this, then if bdist_msi is run programmatically
-        # from within a larger script, subsequent editting of the MSI is blocked.
+        # Cause the MSI file to be released. Without this, then if bdist_msi
+        # is run programmatically from within a larger script, subsequent
+        # editting of the MSI is blocked.
         self.db = None
 
 
